import argparse

from . import local
from . import api
from . import utils


def main():
    parser = argparse.ArgumentParser(
        prog='chromepilot',
        description='Chromedriver download manager and shorcuts for Selenium.',
    )

    subparsers = parser.add_subparsers(dest='command', metavar='')
    subparsers.add_parser(
        'search',
        help='Searches for installs of Google Chrome locally.'
    )
    subparsers.add_parser(
        'check',
        help='Checks the version of Google Chrome and chromedrivers.'
    )

    upgrade = subparsers.add_parser(
        'upgrade',
        help=("Download newer versions of chromedriver. "
              "'upgrade -c' cleans after install.")
    )
    upgrade.add_argument(
        '-c',
        const=1,
        action='store_const'
    )

    subparsers.add_parser(
        'clean',
        help='Searches for outdated chromedrivers locally.'
    )

    subparsers.add_parser(
        'write',
        help='Writes "pilot.toml" template on current directory'
    )

    args = parser.parse_args()

    if args.command == 'search':
        search = local.search_chrome_installed()
        chromes = len(search)
        s = 's' if chromes > 1 else ''
        print(f'{chromes} install{s} could be found.')
        if chromes:
            print('Version:')
        for version in search:
            print(version)

    elif args.command == 'check':
        search = local.search_chrome_installed()
        chromedrivers = local.get_installed_chromedrivers()

        newer_chrome = utils.get_newer_version(search)
        newer_local_chromedriver = utils.get_newer_version(chromedrivers)

        newer_online_chromedriver = api.get_newer_version_info()

        print(
            f'Installs of Google Chrome found: {len(search)}.',
            f'Newer local Google Chrome version found: {newer_chrome}',
            f'Installed chromedrivers: {len(chromedrivers)}.',
            f'Newer local chromedriver: {newer_local_chromedriver}',
            f'Newer chromedriver available online: {newer_online_chromedriver}',  # noqa: E501
            sep='\n'
        )

    elif args.command == 'upgrade':
        newer_online_chromedriver = api.get_newer_version_info()
        chromedrivers = local.get_installed_chromedrivers()
        newer_local_chromedriver = utils.get_newer_version(chromedrivers)

        if utils.is_x_newer_than_y(
            newer_online_chromedriver, newer_local_chromedriver
        ):
            print('A new chromedriver is available online.',
                  'Consider upgrading Google Chrome before '
                  'downloading the chromedriver.',
                  sep='\n')
            answer = input('Install latest chromedriver? [y/N]: ')
            if answer.lower() in ('y', 'yes'):
                print(api.download_newer_version())
                if args.c is not None:
                    local.remove_outdated()
                    print('Cleaned older versions.')
        else:
            print('Everything is up to date.')

    elif args.command == 'clean':
        chromedrivers = local.get_installed_chromedrivers()
        if len(chromedrivers) > 1:
            print('Found more than one chromedriver.')
            answer = input('Delete outdated? [y/N]: ')
            if answer.lower() in ('y', 'yes'):
                local.remove_outdated()
                print('Cleaned older versions.')
        else:
            print('Everything clean.')

    elif args.command == 'write':
        import shutil
        import os
        from pathlib import Path

        if not os.path.exists('pilot.toml'):
            shutil.copy(Path(local.__file__).parent / 'pilot.toml', '.')
        else:
            print("There's already a \"pilot.toml\" in current directory.")

    elif args.command is None:
        parser.print_help()
        print()
        print(
            'In your runtime, try:',
            '',
            'from chromepilot import short',
            'driver = short.driver()  '
            '# Easy access to pre-configured driver',
            sep='\n'
        )



import sys
import os


if sys.platform == 'linux':
    path = '~/.cache/chromepilot/bin/'

elif sys.platform == 'win32':
    path = os.environ['LOCALAPPDATA'] + '\\chromepilot\\bin\\'

os.makedirs(path, exist_ok=True)
INSTALL_PATH = os.path.abspath(path)



import os
import sys
import zipfile
from io import BytesIO
from urllib.request import urlopen

from .__init__ import INSTALL_PATH
from . import local


URL = 'https://chromedriver.storage.googleapis.com/'


def get_newer_version_info() -> str:
    r = urlopen(
        URL + 'LATEST_RELEASE'
    )
    return r.read().decode()


def download_newer_version() -> str:

    version = get_newer_version_info()
    if version in local.get_installed_chromedrivers():
        return 'Latest chromedriver is installed.'

    plat = sys.platform
    if plat == 'linux':
        plat += '64'

    version += '/'
    file = urlopen(
        URL + version + f'chromedriver_{plat}.zip'
    )

    bio = BytesIO(file.read())
    zip_handler = zipfile.ZipFile(bio)
    if (err := zip_handler.testzip()) is None:

        download_path = os.path.join(INSTALL_PATH, version) + os.path.sep
        os.mkdir(download_path)

        driver = zip_handler.open('chromedriver.exe')
        content = driver.read()
        with open(download_path + 'chromedriver.exe', 'wb') as out:
            out.write(content)

        return "Download was successful"

    return f"Error on testzip. File: {err}"



import os
import pathlib
try:
    import tomllib
except ImportError:
    import tomlkit as tomllib

from .__init__ import INSTALL_PATH

HOME = pathlib.Path(INSTALL_PATH).parent


def import_as(x: dict[str, str]) -> None:
    for i, j in x.items():
        exec('import %s as %s' % (i, j))
        exec('globals().update({"%s": %s})' % (j, j))


def from_import(x: dict[str, str]) -> None:
    for i, j in x.items():
        exec('from %s import %s' % (i, j))
        exec('globals().update({"%s": %s})' % (j, j))


def from_import_as(x: dict[str, list[str, str]]):
    for i, j in x.items():
        exec('from %s import %s as %s' % (i, j[0], j[1]))
        exec('globals().update({"%s": %s})' % (j[1], j[1]))


def parse_toml(path):
    with open(path, 'rb') as f:
        file = tomllib.load(f)

    imp_as = file.get('import', {}).get('as')
    from_imp = file.get('from', {}).get('import')
    from_imp_as = from_imp.pop('as', None)
    if imp_as is not None:
        import_as(imp_as)
    if from_imp is not None:
        from_import(from_imp)
    if from_imp_as is not None:
        from_import_as(from_imp_as)

    global parsed_options

    service = file.get('service', {})
    add_argument = file.get('add_argument', {})
    experimental = file.get('add_experimental_option', {})
    imp_wait = file.get('implicitly_wait', {})

    if parsed_options:
        parsed_options['service'].update(service)
        parsed_options['add_argument'].update(add_argument)
        parsed_options['experimental'].update(experimental)
        parsed_options['implicitly_wait'].update(imp_wait)
    else:
        parsed_options['service'] = service
        parsed_options['add_argument'] = add_argument
        parsed_options['experimental'] = experimental
        parsed_options['implicitly_wait'] = imp_wait


def driver(use_toml: bool = True):
    import os

    from selenium import webdriver
    from selenium.webdriver.chrome.service import Service

    from .__init__ import INSTALL_PATH
    from .local import get_installed_chromedrivers
    from .utils import get_newer_version

    chromedrivers = get_installed_chromedrivers()
    if not chromedrivers:
        return
    newer_chromedriver = get_newer_version(chromedrivers)

    path = os.path.join(INSTALL_PATH, newer_chromedriver)
    path += os.path.sep + 'chromedriver.exe'

    if use_toml:
        global parsed_options

        service = parsed_options.get('service', {})
        add_argument = parsed_options.get('add_argument', {})
        experimental = parsed_options.get('experimental', {})
        imp_wait = parsed_options.get('implicitly_wait', {})

    else:
        service, add_argument, experimental, imp_wait = {}, {}, {}, {}

    port = service.get('port', 0)
    service_args = service.get('service_args', None)
    log_path = service.get('log_path', None)

    service = Service(
            executable_path=path,
            port=port,
            service_args=service_args,
            log_path=log_path
        )
    options = webdriver.ChromeOptions()

    args = [key for key in add_argument if add_argument[key]]
    for arg in args:
        options.add_argument(arg)

    for arg in experimental:
        options.add_experimental_option(arg, experimental[arg])

    driver = webdriver.Chrome(service=service, options=options)
    driver.implicitly_wait(imp_wait.get('value', 0))

    return driver


def get_local_toml_path():
    """
    Searches for pilot.toml in the same dir of users main.py
    not os.getcwd()

    """
    import sys

    user_main_file_dir = os.path.abspath(pathlib.Path(sys.argv[0]).parent)
    return os.path.join(user_main_file_dir, 'pilot.toml')


parsed_options = {}

global_toml = os.path.exists(HOME / 'pilot.toml')
if global_toml:
    parse_toml(HOME / 'pilot.toml')

local_toml_path = get_local_toml_path()
local_toml = os.path.exists(local_toml_path)
if local_toml:
    parse_toml(local_toml_path)

del (HOME, INSTALL_PATH, os, tomllib, pathlib,
     parse_toml, import_as, from_import, from_import_as,
     local_toml_path, get_local_toml_path)



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    page_description = f.read()

# Must be full path for build to find
with open('C:\\Users\\henri\\github\\chromepilot\\requirements.txt') as f:
    requirements = f.read()

setup(
    name='chromepilot',
    version='0.0.4',
    author='Henrique do Val',
    author_email='henrique.val@hotmail.com',
    description='Chromedriver download manager and imports shortcuts.',
    long_description=page_description,
    long_description_content_type='text/markdown',
    url='https://github.com/HenriquedoVal/chromepilot',
    packages=find_packages(),
    package_data={'': ['pilot.toml']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'chromepilot = chromepilot.cli:main'
        ]},
    install_requires=requirements,
    python_requires='>=3.8'
)



import os
import sys
import re
from pathlib import Path

from .__init__ import INSTALL_PATH


usual = 'Google\\Chrome\\Application'
X86_PATH = os.path.join(
    os.environ.get('HOMEDRIVE', 'NONE') + '\\',
    'Program Files (x86)',
    usual,
    'chrome.exe'
)

LOCALAPPDATA_PATH = os.path.join(
    os.environ.get('LOCALAPPDATA', 'NONE'),
    usual,
    'chrome.exe'
)

ALL_SEARCH_PATHS = [
    X86_PATH,
    LOCALAPPDATA_PATH
]


def _search_chrome_windows() -> list:
    valid_paths = []
    for path in ALL_SEARCH_PATHS:
        if os.path.exists(path):
            valid_paths.append(path)

    valid_paths = [i.removesuffix('chrome.exe') for i in valid_paths]

    versions = []
    for path in valid_paths:
        for out in os.listdir(path):
            if m := re.match('[0-9]+.0.[0-9]+.[0-9]+', out):
                versions.append(m.string)

    if versions:
        return versions


def _search_chrome_linux() -> list:
    import shutil

    if (bin := shutil.which('google-chrome')) is not None:
        res = os.popen(f'{bin} --version').read()
        return re.findall('[0-9]+.0.[0-9]+.[0-9]+', res)


def search_chrome_installed() -> list:
    if sys.platform == 'win32':
        return _search_chrome_windows()
    elif sys.platform == 'linux':
        return _search_chrome_linux()


def get_installed_chromedrivers() -> list:
    return [i for i in os.listdir(INSTALL_PATH)]


def remove_outdated() -> None:
    from .utils import get_newer_version

    chromedrivers = get_installed_chromedrivers()
    newer = get_newer_version(chromedrivers)
    chromedrivers.remove(newer)

    for directory in chromedrivers:
        os.remove(os.path.join(
            INSTALL_PATH,
            directory,
            'chromedriver.exe'
        ))
        os.rmdir(os.path.join(
            INSTALL_PATH,
            directory
        ))



def get_newer_version(versions: list[str]) -> str:
    versions.sort(
        key=lambda x: (x.split('.')[0],
                       x.split('.')[2],
                       x.split('.')[3]),
        reverse=True
    )

    if len(versions) > 0:
        return versions[0]


def is_x_newer_than_y(x: str, y: str) -> bool:
    if y is None:
        return True

    x = [int(i) for i in x.split('.')]
    y = [int(i) for i in y.split('.')]
    for ind in (0, 2, 3):
        if x[ind] > y[ind]:
            return True
    return False
